import asyncio
import functools

from async_property.proxy import AwaitableOnly

is_coroutine = asyncio.iscoroutinefunction


def async_property(func, *args, **kwargs):
    assert is_coroutine(func), 'Can only use with async def'
    return AsyncPropertyDescriptor(func, *args, **kwargs)


class AsyncPropertyDescriptor:
    def __init__(self, _fget, field_name=None):
        self._fget = _fget
        self.field_name = field_name or _fget.__name__
        functools.update_wrapper(self, _fget)

    def __set_name__(self, owner, name):
        self.field_name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.awaitable_only(instance)

    def __set__(self, instance, value):
        raise ValueError(INVALID_ACTION.format('set'))

    def __delete__(self, instance):
        raise ValueError(INVALID_ACTION.format('delete'))

    def get_loader(self, instance):
        @functools.wraps(self._fget)
        async def get_value():
            return await self._fget(instance)
        return get_value

    def awaitable_only(self, instance):
        return AwaitableOnly(self.get_loader(instance))


INVALID_ACTION = 'Cannot {} @async_property. ' \
                 'Use @async_cached_property instead.'



import asyncio
from async_property.cached import AsyncCachedPropertyDescriptor

is_coroutine = asyncio.iscoroutinefunction


AWAIT_LOADER_ATTR = '_async_property_loaders'


def get_loaders(instance):
    return getattr(instance, AWAIT_LOADER_ATTR, ())


class AwaitLoaderMeta(type):
    def __new__(mcs, name, bases, attrs) -> type:
        loaders = {}
        for key, value in attrs.items():
            if isinstance(value, AsyncCachedPropertyDescriptor):
                loaders[key] = value.get_loader

        for base in reversed(bases):
            for field, get_loader in get_loaders(base):
                if field not in loaders:
                    loaders[field] = get_loader

        attrs[AWAIT_LOADER_ATTR] = tuple(loaders.items())
        return super().__new__(mcs, name, bases, attrs)


class AwaitLoader(metaclass=AwaitLoaderMeta):
    def __await__(self):
        return self._load().__await__()

    async def _load(self):
        """
        Calls overridable async load method
        and then calls async property loaders
        """
        if hasattr(self, 'load') and is_coroutine(self.load):
            await self.load()
        loaders = get_loaders(self)
        if loaders:
            await asyncio.wait([
                asyncio.create_task(get_loader(self)())
                for field, get_loader
                in loaders
            ])
        return self



import asyncio
import time

import pytest

from async_property import async_cached_property, AwaitLoader

pytestmark = pytest.mark.asyncio


class MyModel(AwaitLoader):
    def __init__(self):
        self.num_loaded = 0

    @async_cached_property
    async def first(self):
        await asyncio.sleep(0.1)
        self.num_loaded += 1

    @async_cached_property
    async def second(self):
        await asyncio.sleep(0.2)
        self.num_loaded += 1

    @async_cached_property
    async def third(self):
        await asyncio.sleep(0.3)
        self.num_loaded += 1


async def test_is_locking():
    start = time.time()
    instance = MyModel()
    await asyncio.gather(instance.first, instance.first, instance.first)
    duration = time.time() - start
    assert instance.num_loaded == 1
    assert 0.1 <= duration < 0.15


async def test_lock_multiple_instances():
    start = time.time()
    instance_one = MyModel()
    instance_two = MyModel()
    await asyncio.gather(instance_one.first, instance_two.first)
    duration = time.time() - start
    assert 0.1 <= duration < 0.15


async def test_concurrent():
    start = time.time()
    instance = MyModel()
    await asyncio.gather(instance.first, instance.second, instance.third,
                         instance.first, instance.second, instance.third)
    duration = time.time() - start
    assert instance.num_loaded == 3
    assert 0.3 <= duration < 0.4


async def test_await_concurrent():
    start = time.time()
    instance = await MyModel()
    duration = time.time() - start
    assert instance.num_loaded == 3
    assert 0.3 <= duration < 0.4



import pytest
from async_property import async_cached_property, AwaitLoader

pytestmark = pytest.mark.asyncio


class MyModel(AwaitLoader):
    @async_cached_property
    async def foo(self):
        return 'bar'


async def test_empty_instance():
    assert await AwaitLoader()


async def test_loader():
    instance = await MyModel()
    assert 'foo' in instance.__async_property__.cache
    assert instance.foo == 'bar'


async def test_delayed_field():
    instance = MyModel()
    coro = instance.foo
    await instance
    assert instance.foo == 'bar'
    assert await coro == 'bar'


async def test_delayed_field_with_setter():
    instance = MyModel()
    coro = instance.foo
    await instance
    assert instance.foo == 'bar'
    instance.foo = 'abc'
    assert await coro == 'abc'


async def test_empty_instance_loaders():
    assert AwaitLoader._async_property_loaders == ()


async def test_instance_loaders():
    assert MyModel._async_property_loaders == (
        ('foo', MyModel.foo.get_loader),
    )


class MyModelWithLoad(AwaitLoader):
    async def load(self):
        self.loaded = True


async def test_call_load():
    instance = await MyModelWithLoad()
    assert instance.loaded == True



import asyncio
import functools
from collections import defaultdict

from async_property.proxy import AwaitableOnly, AwaitableProxy

is_coroutine = asyncio.iscoroutinefunction


ASYNC_PROPERTY_ATTR = '__async_property__'


def async_cached_property(func, *args, **kwargs):
    assert is_coroutine(func), 'Can only use with async def'
    return AsyncCachedPropertyDescriptor(func, *args, **kwargs)


class AsyncCachedPropertyInstanceState:
    def __init__(self):
        self.cache = {}
        self.lock = defaultdict(asyncio.Lock)

    __slots__ = 'cache', 'lock'


class AsyncCachedPropertyDescriptor:
    def __init__(self, _fget, _fset=None, _fdel=None, field_name=None):
        self._fget = _fget
        self._fset = _fset
        self._fdel = _fdel
        self.field_name = field_name or _fget.__name__

        functools.update_wrapper(self, _fget)
        self._check_method_sync(_fset, 'setter')
        self._check_method_sync(_fdel, 'deleter')

    def __set_name__(self, owner, name):
        self.field_name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if self.has_cache_value(instance):
            return self.already_loaded(instance)
        return self.not_loaded(instance)

    def __set__(self, instance, value):
        if self._fset is not None:
            self._fset(instance, value)
        self.set_cache_value(instance, value)

    def __delete__(self, instance):
        if self._fdel is not None:
            self._fdel(instance)
        self.del_cache_value(instance)

    def setter(self, method):
        self._check_method_name(method, 'setter')
        return type(self)(self._fget, method, self._fdel, self.field_name)

    def deleter(self, method):
        self._check_method_name(method, 'deleter')
        return type(self)(self._fget, self._fset, method, self.field_name)

    def _check_method_name(self, method, method_type):
        if method.__name__ != self.field_name:
            raise AssertionError(
                f'@{self.field_name}.{method_type} name must match property name'
            )

    def _check_method_sync(self, method, method_type):
        if method and is_coroutine(method):
            raise AssertionError(
                f'@{self.field_name}.{method_type} must be synchronous'
            )

    def get_instance_state(self, instance):
        try:
            return getattr(instance, ASYNC_PROPERTY_ATTR)
        except AttributeError:
            state = AsyncCachedPropertyInstanceState()
            object.__setattr__(instance, ASYNC_PROPERTY_ATTR, state)
            return state

    def get_lock(self, instance):
        lock = self.get_instance_state(instance).lock
        return lock[self.field_name]

    def get_cache(self, instance):
        return self.get_instance_state(instance).cache

    def has_cache_value(self, instance):
        cache = self.get_cache(instance)
        return self.field_name in cache

    def get_cache_value(self, instance):
        cache = self.get_cache(instance)
        return cache[self.field_name]

    def set_cache_value(self, instance, value):
        cache = self.get_cache(instance)
        cache[self.field_name] = value

    def del_cache_value(self, instance):
        cache = self.get_cache(instance)
        del cache[self.field_name]

    def get_loader(self, instance):
        @functools.wraps(self._fget)
        async def load_value():
            async with self.get_lock(instance):
                if self.has_cache_value(instance):
                    return self.get_cache_value(instance)
                value = await self._fget(instance)
                self.__set__(instance, value)
                return value
        return load_value

    def already_loaded(self, instance):
        return AwaitableProxy(self.get_cache_value(instance))

    def not_loaded(self, instance):
        return AwaitableOnly(self.get_loader(instance))



__author__ = """Ryan Anguiano"""
__email__ = 'ryan.anguiano@gmail.com'
__version__ = '0.2.2'


from .base import async_property
from .cached import async_cached_property
from .loader import AwaitLoader
from .proxy import AwaitableOnly


__all__ = ['async_property', 'async_cached_property', 'AwaitLoader', 'AwaitableOnly']



import pytest

from async_property import async_property
from async_property.base import AsyncPropertyDescriptor
from async_property.proxy import AwaitableOnly

pytestmark = pytest.mark.asyncio


class MyModel:
    @async_property
    async def foo(self) -> str:
        return 'bar'


async def test_descriptor():
    assert isinstance(MyModel.foo, AsyncPropertyDescriptor)
    assert MyModel.foo.__name__ == 'foo'
    assert MyModel.foo.__annotations__['return'] == str


async def test_property():
    instance = MyModel()
    assert isinstance(instance.foo, AwaitableOnly)
    assert await instance.foo == 'bar'


async def test_multiple_calls():
    instance = MyModel()
    assert await instance.foo == 'bar'
    assert await instance.foo == 'bar'


async def test_setter():
    instance = MyModel()
    with pytest.raises(ValueError):
        instance.foo = 'abc'


async def test_deleter():
    instance = MyModel()
    with pytest.raises(ValueError):
        del instance.foo


async def test_sync_error():
    with pytest.raises(AssertionError):
        class BadModel:
            @async_property
            def foo(self):
                return 'bar'



import pytest

from async_property import async_cached_property, AwaitLoader
from async_property.loader import get_loaders

pytestmark = pytest.mark.asyncio


class TestLoaderA(AwaitLoader):
    @async_cached_property
    async def foo(self):
        return True


class TestLoaderB(AwaitLoader):
    @async_cached_property
    async def bar(self):
        return False


class TestLoader(TestLoaderA, TestLoaderB):
    @async_cached_property
    async def abc(self):
        return 123

    @async_cached_property
    async def foo(self):
        return False


async def test_loaders_exist():
    instance = TestLoader()
    assert len(get_loaders(instance)) == 3


async def test_loader_keys():
    instance = TestLoader()
    assert set(dict(get_loaders(instance))) == {'foo', 'bar', 'abc'}


async def test_inherited_value():
    instance = await TestLoader()
    assert instance.foo == False

    base = await TestLoaderA()
    assert base.foo == True



#!/usr/bin/env python
"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = []

setup_requirements = ['pytest-runner']

test_requirements = ['pytest']

setup(
    author="Ryan Anguiano",
    author_email='ryan.anguiano@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development',
        'Topic :: Utilities',
    ],
    description="Python decorator for async properties.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='async_property',
    name='async_property',
    packages=find_packages(include=['async_property']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/ryananguiano/async_property',
    version='0.2.2',
    zip_safe=False,
)



import pytest

from async_property import async_cached_property
from async_property.cached import AsyncCachedPropertyDescriptor
from async_property.proxy import AwaitableOnly, AwaitableProxy

pytestmark = pytest.mark.asyncio


class MyModel:
    @async_cached_property
    async def foo(self) -> str:
        return 'bar'


async def test_descriptor():
    assert isinstance(MyModel.foo, AsyncCachedPropertyDescriptor)
    assert MyModel.foo.__name__ == 'foo'
    assert MyModel.foo.__annotations__['return'] == str


async def test_field():
    instance = MyModel()
    assert isinstance(instance.foo, AwaitableOnly)
    assert await instance.foo == 'bar'
    assert 'foo' in instance.__async_property__.cache


async def test_awaited_repeated():
    instance = MyModel()
    assert await instance.foo == 'bar'
    assert isinstance(instance.foo, AwaitableProxy)
    assert instance.foo == 'bar'
    assert await instance.foo == 'bar'


async def test_default_setter():
    instance = MyModel()
    instance.foo = 'abc'
    assert 'foo' in instance.__async_property__.cache
    assert instance.foo == 'abc'


async def test_default_deleter():
    instance = MyModel()
    await instance.foo
    assert 'foo' in instance.__async_property__.cache
    del instance.foo
    assert 'foo' not in instance.__async_property__.cache


class ModelWithSetterDeleter:
    @async_cached_property
    async def foo(self):
        return 'bar'

    @foo.setter
    def foo(self, value):
        self.bar = '123'

    @foo.deleter
    def foo(self):
        del self.bar


async def test_async_property_with_setter():
    instance = ModelWithSetterDeleter()
    instance.foo = 'abc'
    assert instance.foo == 'abc'
    assert await instance.foo == 'abc'
    assert 'foo' in instance.__async_property__.cache
    assert hasattr(instance, 'bar')
    assert instance.bar == '123'


async def test_async_property_with_deleter():
    instance = ModelWithSetterDeleter()
    await instance.foo
    assert 'foo' in instance.__async_property__.cache
    assert hasattr(instance, 'bar')
    del instance.foo
    assert 'foo' not in instance.__async_property__.cache
    assert not hasattr(instance, 'bar')


class MyModelWithMultiple:
    @async_cached_property
    async def first(self):
        return 123

    @async_cached_property
    async def second(self):
        return 456


async def test_multiple_fields():
    instance = MyModelWithMultiple()
    assert await instance.first == 123
    assert await instance.second == 456


async def test_bad_setter_name():
    with pytest.raises(AssertionError):
        class BadSetter:
            @async_cached_property
            async def foo(self):
                return True

            @foo.setter
            def not_foo(self, value):
                pass


async def test_async_setter():
    with pytest.raises(AssertionError):
        class AsyncSetter:
            @async_cached_property
            async def foo(self):
                return True

            @foo.setter
            async def foo(self, value):
                pass
